# Documentation from development
This documentation captures some of the "necessary things" that have been encountered during development and how this extension has become what it is. It's intended for people that want to continue developing the extension or make a similar one or just to help with some aspects of developing a VSC extension.  
A useful documentation for the package.json seems to be [this](https://docs.npmjs.com/files/package.json).

## The beginning
The whole thing started as a "simple" syntax highlighting extension. The basic structure was provided by using a Yeoman template for "Language Support" ([see for more details](https://code.visualstudio.com/api/language-extensions/syntax-highlight-guide) including links to some useful tutorials and resources). As such it used TextMate (... code?) for syntax highlighting.  
It's good to note that the syntax highlighting can be split into several languages which then reference each other, or rather in this case the "LEO" language, which is then referenced by the AdoScript language at the relevant places. While AdoScript seems to be a "Specialization" of the "LEO" language it also means that it can't simply reference the top element of the LEO language and be done. Instead it can only reuse certain parts, like the "bracket-expression" (through `"include": "source.leo#bracket-expression"`) or "comments" (through `"include": "source.leo#comment"`) which are often encountered in AdoScript. However since many of the other things are restricted in AdoScript (e.g. what keywords are available) it can't simply reference the top level of LEO, but instead reuses the specific parts that fit. Others had to be tailored into AdoScript specifically (e.g. the parameter with modifier style for the `LEO` command, because it doesn't seem to be used in many other places of AdoScript), which unfortunately means that some changes have to be synchronized between the two languages to keep the entire things consistent. Additionally AdoScript seems to also extend the LEO-grammar not (i.e. it's not simply "inheritance by restriction" but also "by extension") through the CC command, which has a string following it and in some cases another main parameter. To be able to reference the other language like that it is necessary to say that one language embeds another in the package.json, through something like this:
```json
{
  "grammars": [
    {
        "language": "adoxx-leo-program",
        "scopeName": "source.leo",
        "path": "./syntaxes/leo.tmLanguage.json"
    },
    {
        "language": "adoxx-ado-script",
        "scopeName": "source.asc",
        "path": "./syntaxes/asc.tmLanguage.json",
        "embeddedLanguages": {
            "meta.embedded.block.leo": "source.leo"
        }
    }
  ],
}
```

## Publishing the first version
In order to properly publish the extension on the marketplace a few more steps were necessary. Basic information about that can be found [here](https://code.visualstudio.com/api/working-with-extensions/publishing-extension).

### Relevant parts for the package.json
The following keys were relevant in the package.json for publishing to provide additional information:
```json
{
  "description": "...", // providing a short description
  "icon": "images/AdoScriptExtension.png", // relative path to the image
  "author": { "name": "Patrik Burzynski" }, // me :D
  "contributors": [{ "name": "Wilfrid Utz" }], // others that helped/contributed
  "publisher": "...", // the publisher id
  "license": "...", // the license, which could additionally be provided through a `LICENSE` file. For example the SPDX license identifier.
  "preview": true, // especially at the beginning
  "version": "...", // the version of the extension
  "engines": { "vscode": "^1.33.0" }, // which VSC versions are supported (1.33 and up)
  "repository": { // where the source code can be found publicly
    "type": "git",
    "url": "https://git.boc-group.eu/adoxx/adoscript-vs-code-support"
  },
  "bugs": { // where bugs can be reported
    "url": "https://git.boc-group.eu/adoxx/adoscript-vs-code-support/issues"
  },
}
```
This means that we had to create and add an icon file, decide on a license to publish under as well as provide the repository for the code and bug-tracking.

### Fixing the README
Since the README is used in several places, including the gitlab repository and the VSC marketplace, I wanted it to be nice and useful for the users. So in the first version there was a Table of Contents at the top that linked to the major sections and the heading contained a link that pointed back to the ToC through an &uarr;. Unfortunately neither gitlab nor the VSC Marketplace liked those links so I had to change the links to fit with the anchors created/used by those sites. This means that the ToC had to have links that:
1. start with a #
1. used - instead of spaces
1. the text of the anchor was lower case
1. no weird characters like &uarr;

here is an example:
```markdown
* [Recommendations and Restrictions - short version](#the-short-version-of-recommendations-and-restrictions)
* [Recommendations - detailed](#recommendations)
* [Restrictions - detailed](#restrictions)
```
As such I had to move the "back to ToC" link to some other place, with some tests I decided to put those at the end of the section, since those can be easily reached when jumping to a section heading by scrolling up a bit and it doesn't put a gap between the heading and the content.

### Creating the package / publishing
To create the package / publish it was necessary to create an Azure DevOps organisation and get a Personal Access Token. Details on that can be found [here](https://code.visualstudio.com/api/working-with-extensions/publishing-extension)  
After that creating the package basically meant installing vsce (I think `npm install vsce` did the trick) and then running `vsce package` or `vsce publish` in the extensions folder. During the packaging / publishing it asked for the Personal Access Token, but that was just a copy and paste kind of deal.

## Further adding "Programmatic features"
In order to add programmatic features (commands through the command palette etc.) I have decided to use TypeScript, because that is also what has been used in the example hello-world project from VSC (obtained using yeoman, [details](https://code.visualstudio.com/api/get-started/your-first-extension)).

### Adding a src directory and copying other relevant configurations
I added a src directory where the `extension.ts` is located (I guess the heart of the programmatic extension?). Additionally I copied the `tsconfig.json` and `tslint.json` from the hello-world project so that I guess TypeScript properly works. `tsconfig.json` might however also be created using a command like `tsc --init`. Also copied a `tasks.json` from the other project, but it seems like it's possible to simply create your own for npm (see [this](https://code.visualstudio.com/docs/editor/tasks) for more info).

### Necessary extension to package.json
The following parts were added to the package.json to a) use Node.js/TypeScript and b) download some basic dependencies for development. The "postinstall" part is apparently also important for installing a dependency when the extension is published.
```json
{
  "scripts": {
    "vscode:prepublish": "npm run compile",
    "compile": "tsc -p ./",
    "watch": "tsc -watch -p ./",
    "postinstall": "node ./node_modules/vscode/bin/install",
    "test": "npm run compile && node ./node_modules/vscode/bin/test"
  },
  "devDependencies": {
    "typescript": "^3.3.1",
    "vscode": "^1.1.28",
    "tslint": "^5.12.1",
    "@types/node": "^10.12.21",
    "@types/mocha": "^2.2.42"
  },
}
```
With those changes I ran `npm install` in the folder where the package.json is located and it downloaded the necessary stuff and installed it (Note: before I did that I actually just ran `npm install vscode` which installed only the vscode dependency, but I'm guessing adding the "devDependencies" helped the `npm install` figure out what is necessary). I think `tslint` and `@types/mocha` aren't really necessary for the simple "hello world" type of extension though. `vscode` and `@types/node` were necessary to run the basic hello world extension and `typescript` was necessary to create the .vsix package.  
The "main" and "commands" (inside the "contributes") were added for simple testing at first of commands through he command palette.
```json
{
  "main": "./out/extension.js",
  "contributes": {
    "commands": [
      {
        "command": "extension.helloWorld",
        "title": "Hello World 2 - hello Worlder"
      }
    ],
  }
}
```