import * as vscode from 'vscode';
import * as path from 'path';

import {runSymbolProvider, insertString} from './extension';

// manages query editor webview panels
export default class InterfacePanel {

	// track the currently panel. only allow a single panel to exist at a time
	public static currentPanel: InterfacePanel | undefined;

	public static readonly viewType = 'queryEditor';

	public variables?: vscode.DocumentSymbol[] = [];

	private readonly _panel: vscode.WebviewPanel;
	private readonly _extensionPath: string;
	private _disposables: vscode.Disposable[] = [];

	public static createOrShow(extensionPath: string, text?: string): void {
		const column = vscode.window.activeTextEditor
			? vscode.window.activeTextEditor.viewColumn
			: undefined;

		// if we already have a panel, show it.
		if (InterfacePanel.currentPanel) {
			InterfacePanel.currentPanel._panel.reveal(column);
			return;
		}

		// otherwise, create a new panel.
		const panel = vscode.window.createWebviewPanel(
			InterfacePanel.viewType,
			'Query Editor',
			vscode.ViewColumn.Beside,
			{
				// enable javascript in the webview
				enableScripts: true,

				// and restrict the webview to only loading content from our extension's resources directory.
				localResourceRoots: [
					vscode.Uri.file(path.join(extensionPath, 'resources')),
					vscode.Uri.file(path.join(extensionPath, 'vendor'))
				]
			}
		);

		InterfacePanel.currentPanel = new InterfacePanel(panel, extensionPath, text);
	}

	public static revive(panel: vscode.WebviewPanel, extensionPath: string): void {
		InterfacePanel.currentPanel = new InterfacePanel(panel, extensionPath);
	}

	private constructor(panel: vscode.WebviewPanel, extensionPath: string, text?: string) {
		this._panel = panel;
		this._extensionPath = extensionPath;

		// set the webview's initial html content
		this._update(text);

		// listen for when the panel is disposed
		// this happens when the user closes the panel or when the panel is closed programatically
		this._panel.onDidDispose(() => this.dispose(), null, this._disposables);

		// handle messages from the webview
		this._panel.webview.onDidReceiveMessage(
			message => {
				switch (message.command) {
					case 'get_data':
						this.sendVariables();
						return;
					case 'send_string':
						insertString(message.string);
						return;
				}
			},
			null,
			this._disposables
		);
	}

	public sendVariables(): void {
		// refresh list of variables
		runSymbolProvider().then(variables => {
			this.variables = variables;

			// send a message to the webview webview
			this._panel.webview.postMessage(
				{
					command: 'post_data',				// here we send the list of parsed variables from the file
					data: this.variables
				}
			);
		});
	}

	public dispose(): void {
		InterfacePanel.currentPanel = undefined;

		// clean up our resources
		this._panel.dispose();

		while (this._disposables.length) {
			const x = this._disposables.pop();
			if (x) {
				x.dispose();
			}
		}
	}

	private _update(text?: string): void {
		const webview = this._panel.webview;

		this._panel.title = "Query Editor";

		if (text !== undefined) {
			this._panel.webview.html = this._getHtmlForWebview(webview, text);
		} else {
			this._panel.webview.html = this._getHtmlForWebview(webview);
		}
	}

	private _getHtmlForWebview(webview: vscode.Webview, text?: string): string {
		// local path relative to main script run in the webview
		const scriptPathOnDisk = vscode.Uri.file(
			path.join(this._extensionPath, 'resources', 'main.js')
		);
		const stylePathOnDisk = vscode.Uri.file(
			path.join(this._extensionPath, 'resources', 'style.css')
		);
		const CodePathOnDisk = vscode.Uri.file(
			path.join(this._extensionPath, 'vendor', 'codemirror', 'lib', 'codemirror.js')
		);
		const CodeStylePathOnDisk = vscode.Uri.file(
			path.join(this._extensionPath, 'vendor', 'codemirror', 'lib', 'codemirror.css')
		);
		const simplePathOnDisk = vscode.Uri.file(
			path.join(this._extensionPath, 'vendor', 'codemirror', 'addon', 'mode', 'simple.js')
		);
		const themePathOnDisk = vscode.Uri.file(
			path.join(this._extensionPath, 'vendor', 'codemirror', 'theme', 'dracula.css')
		);
		const closebracketPathOnDisk = vscode.Uri.file(
			path.join(this._extensionPath, 'vendor', 'codemirror', 'addon', 'edit', 'closebrackets.js')
		);

		// and the uri we use to load this script in the webview
		const scriptUri = webview.asWebviewUri(scriptPathOnDisk);
		const styleUri = webview.asWebviewUri(stylePathOnDisk);
		const codeMirrorJsUri = webview.asWebviewUri(CodePathOnDisk);
		const codeMirrorCssUri = webview.asWebviewUri(CodeStylePathOnDisk);
		const simpleUri = webview.asWebviewUri(simplePathOnDisk);
		const themeUri = webview.asWebviewUri(themePathOnDisk);
		const closebracketUri = webview.asWebviewUri(closebracketPathOnDisk);

		// use a nonce to whitelist which scripts can be run
		const nonce = getNonce();

		return `<!DOCTYPE html>
            <html lang="en">
            <head>
                <meta charset="UTF-8">

                <!--
                Use a content security policy to only allow loading images from https or from our extension directory,
                and only allow scripts that have a specific nonce.
				-->
				<meta http-equiv="Content-Security-Policy" content="default-src 'none'; img-src ${webview.cspSource} https:; script-src 'nonce-${nonce}';style-src ${webview.cspSource};">

                <meta name="viewport" content="width=device-width, initial-scale=1.0">
				<title>Query Editor</title>

				<script nonce="${nonce}" src="${codeMirrorJsUri}"></script>
				<script nonce="${nonce}" src="${closebracketUri}"></script>
				<script nonce="${nonce}" src="${simpleUri}"></script>
				<link rel="stylesheet" href="${codeMirrorCssUri}">
				<link rel="stylesheet" href="${themeUri}">
				
				<link rel="stylesheet" href="${styleUri}">
				</head>
				
				<body>
				<br/>
                <!--<img src="" width="300" />-->
				<!--<h1 id="lines-of-code-counter">0</h1>-->
				
				<label for="text_query">Type your AQL query:</label>
				
				<br/>
				
				<textarea name="Text Query" id="text_query" rows="4" cols="80">${text ? toQuery(text) : ""}</textarea>
				
				<br/>
				<br/>
				
				<div class="btn_container">
					<button type="button" id="btn_or">OR</button>
					<button type="button" id="btn_and">AND</button>
					<button type="button" id="btn_not">NOT</button>
					<button type="button" id="btn_diff">DIFF</button>
					<button type="button" id="btn_like">LIKE</button>
				</div>

				<br/>

				<div class="btn_container">
					<div id="div_class">
						<button type="button" id="btn_class">Insert CLASS:</button>
						<input name="text" list="options_class" id="sel_class">
							<datalist id="options_class"></datalist>
						</input>
					</div>
				</div>

				<br/>

				<div class="btn_container">
					<button type="button" id="btn_show">Show/Hide Variabes</button>
					<button type="button" id="btn_generate">Generate AQL String</button>
					<button type="button" id="btn_insert">Insert AQL</button>
				</div>

				<br/>

				<textarea id="text_result" name="text_query" rows="4" cols="80"></textarea>

				<br/>
				
				<div id="var_ctn">
				</div>
				
				<script nonce="${nonce}" src="${scriptUri}"></script>
            </body>
            </html>`;
	}
}

function getNonce(): string {
	let text = '';
	const possible = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789';
	for (let i = 0; i < 32; i++) {
		text += possible.charAt(Math.floor(Math.random() * possible.length));
	}
	return text;
}

function toQuery(text: string): string {
	let ret = text;

	ret = ret.replace(/\\"\"\s\+\s(.+?)\s\+\s\"\\"/g, "$${$1}$$");
	ret = ret.replace(/\\\"/g, "\"");
	ret = ret.replace(/(^"|"$)/g, "");

	return ret;
}
